from __future__ import print_function
import yaml
from lymph.cli.base import Command


class ConfigCommand(Command):
    """
    Usage: lymph config [options]

    Prints configuration for inspection

    {COMMON_OPTIONS}
    """

    short_description = 'Prints configuration for inspection'

    def run(self):
        print(yaml.safe_dump(self.config.values))



# -*- coding: utf-8 -*-
from __future__ import print_function

from lymph.client import Client
from lymph.cli.base import Command, handle_request_errors


class LogLevelCommand(Command):
    """
    Usage: lymph change-loglevel <address> [options]

    Sets the log level of a service's logger (all instances) for a given amount of time and then resets.

    Options:
      --name=<name>, -n            Logger name to change.
      --level=<level>, -l          Logging level to use.
      --period=<seconds>, -p       Period where change will be effected, after
                                   the logging level will be reverted [default: 60]
      --guess-external-ip, -g      Guess the public facing IP of this machine and
                                   use it instead of the provided address.

    {COMMON_OPTIONS}
    """

    short_description = 'Set logging level of a service logger'

    @handle_request_errors
    def run(self):
        try:
            period = float(self.args.get('--period'))
        except ValueError:
            print("--period requires a number (e.g. --period=0.42)")
            return 1

        return self.change_loglevel(
            address=self.args['<address>'],
            logger=self.args['--name'],
            level=self.args['--level'],
            period=period,
        )

    def change_loglevel(self, address, logger, level, period):
        client = Client.from_config(self.config)
        body = {
            'qualname': logger,
            'loglevel': level,
            'period': period,
        }
        service = client.container.lookup(address)
        print("Changing logger '%s' of '%s' to '%s' for a period of %s seconds" %
              (logger, address, level, period))
        for instance in service:
            client.request(instance.endpoint, 'lymph.change_loglevel', body)




from __future__ import absolute_import

import sys
from contextlib import contextmanager
from collections import namedtuple

import logging
import kombu
import kombu.mixins
import kombu.pools

from lymph.events.base import BaseEventSystem
from lymph.core.events import Event
from lymph.utils.logging import setup_logger


logger = logging.getLogger(__name__)


DEFAULT_SERIALIZER = 'lymph-msgpack'
DEFAULT_EXCHANGE = 'lymph'
DEFAULT_MAX_RETRIES = 3
RETRY_HEADER = 'retry_count'

# Info of where the queue master server, so that we can redeclare the queue
# when we failover to another master.
QueueInfo = namedtuple('QueueInfo', 'master queue')


class EventConsumer(kombu.mixins.ConsumerMixin):
    def __init__(self, event_system, connection, queue, handler, exchange, max_retries=DEFAULT_MAX_RETRIES):
        self.connection = connection
        self.queue = queue
        self.handler = handler
        self.exchange = exchange
        self.greenlet = None
        self.event_system = event_system
        self.connect_max_retries = max_retries
        self.retry_producer = EventProducer(
            exchange=self.event_system.retry_exchange,
            routing_key=handler.queue_name,
            event_system=event_system,
        )

    def get_consumers(self, Consumer, channel):
        return [Consumer(queues=[self.queue], callbacks=[self.on_kombu_message])]

    def create_connection(self):
        return kombu.pools.connections[self.connection].acquire()

    def on_connection_revived(self):
        # Re-create any queue and bind it to exchange in case of a failover, Since
        # the bind may be broken or the queue my have been deleted if it was created
        # with auto-delete.
        self._declare()

    def _declare(self):
        with self.establish_connection() as conn:
            self.event_system.safe_declare(conn, self.queue)
            for event_type in self.handler.event_types:
                self.queue(conn).bind_to(exchange=self.exchange, routing_key=event_type)
                self.queue(conn).bind_to(exchange=self.event_system.retry_exchange, routing_key=self.handler.queue_name)

    def on_kombu_message(self, body, message):
        logger.debug("received kombu message %r", body)
        if self.handler.sequential:
            self._handle_message(body, message)
        else:
            self.event_system.container.spawn(self._handle_message, body, message)

    def _handle_message(self, body, message):
        try:
            event = Event.deserialize(body)
            self.handler(event)
            message.ack()
        except:
            self._handle_fail(message, event)
        finally:
            if self.handler.once:
                self.event_system.unsubscribe(self.handler)

    def _handle_fail(self, message, event):
        retry = event.headers.get(RETRY_HEADER, self.handler.retry) - 1
        if retry >= 0:
            self._requeue(event, retry)
        message.reject()
        logger.exception('failed to handle event from queue %r (tries_left=%s)', self.handler.queue_name, retry)
        self.event_system.container.error_hook(sys.exc_info())

    def _requeue(self, event, retry):
        event.headers[RETRY_HEADER] = retry
        self.retry_producer.emit(event)

    def start(self):
        if self.greenlet:
            return
        self.should_stop = False
        self._declare()
        self.greenlet = self.event_system.container.spawn(self.run)

    def stop(self, **kwargs):
        if not self.greenlet:
            return
        self.should_stop = True
        self.greenlet.join()
        self.greenlet = None


class EventProducer(object):
    def __init__(self, exchange, routing_key, event_system, max_retries=DEFAULT_MAX_RETRIES):
        self.routing_key = routing_key
        self.exchange = exchange
        self.event_system = event_system
        self.max_retries = max_retries

    @contextmanager
    def _get_connection(self):
        with self.event_system.get_connection() as conn:
            yield conn

    def _get_producer(self, conn):
        return conn.Producer(
            serializer=self.event_system.serializer, routing_key=self.routing_key,
            exchange=self.exchange)

    def emit(self, event):
        with self._get_connection() as conn:
            producer = self._get_producer(conn)
            return producer.publish(event.serialize(), retry_policy={'max_retries': self.max_retries})


class EventProducerWithDelay(EventProducer):
    """Producer that allow sending messages after a given delay.

    It works by publishing messages to an intermediate RabbitMQ queue, this messages
    will have a ttl set to the delay given, this way RabbitMQ will forward the messages
    after the ttl expire to the dead-letter-exchange attached to the intermediate queue,
    which we set to our main exchange (i.e. default to 'lymph' exchange), et voila now
    RabbitMQ can send the message to the lymph events handler.
    """

    def __init__(self, delay, *args, **kwargs):
        super(EventProducerWithDelay, self).__init__(*args, **kwargs)
        self.delay = delay  # Delay in ms.
        self._intermediate_queue = None

    def _get_producer(self, conn):
        if self._intermediate_queue is None or self._intermediate_queue.master != conn.as_uri():
            queue = self._prepare_intermediate_queue(conn)
            self._intermediate_queue = QueueInfo(master=conn.as_uri(), queue=queue)
        return super(EventProducerWithDelay, self)._get_producer(conn)

    def _prepare_intermediate_queue(self, conn):
        queue_name = '%s-wait_%s' % (self.routing_key, self.delay)
        queue = self.event_system.get_queue(queue_name, durable=False, queue_arguments={
            'x-dead-letter-exchange': self.event_system.exchange.name,
            'x-dead-letter-routing-key': self.routing_key,
            'x-message-ttl': self.delay,
        })
        self.event_system.safe_declare(conn, queue)
        queue(conn).bind_to(exchange=self.exchange, routing_key=self.routing_key)
        return queue


class KombuEventSystem(BaseEventSystem):
    def __init__(self, connection, exchange_name, serializer=DEFAULT_SERIALIZER, connect_max_retries=DEFAULT_MAX_RETRIES):
        super(KombuEventSystem, self).__init__()
        self.connection = connection
        self.exchange_name = exchange_name
        self.exchange = kombu.Exchange(exchange_name, 'topic', durable=True)
        self.waiting_exchange = kombu.Exchange('%s_waiting' % exchange_name, 'direct', durable=True)
        self.retry_exchange = kombu.Exchange('%s_retry' % exchange_name, 'direct', durable=True)
        self.serializer = serializer
        self.connect_max_retries = connect_max_retries
        self._producers = {}
        self.consumers_by_queue = {}

    @classmethod
    def from_config(cls, config, **kwargs):
        exchange_name = config.get('exchange', DEFAULT_EXCHANGE)
        serializer = config.get('serializer', DEFAULT_SERIALIZER)
        connection = kombu.Connection(**config)
        return cls(connection, exchange_name, serializer=serializer, **kwargs)

    def on_start(self):
        setup_logger('kombu')
        with self.get_connection() as conn:
            self.exchange(conn).declare()
            self.waiting_exchange(conn).declare()
            self.retry_exchange(conn).declare()

    def on_stop(self, **kwargs):
        for consumer in self.consumers_by_queue.values():
            consumer.stop(**kwargs)
        self.consumers_by_queue.clear()

    def subscribe(self, handler, consume=True):
        try:
            consumer = self.consumers_by_queue[handler.queue_name]
        except KeyError:
            consumer = self.setup_consumer(handler)
        else:
            if consumer.handler != handler:
                raise RuntimeError('cannot subscribe to queue %r more than once' % handler.queue_name)
        if consume:
            consumer.start()
        return consumer

    def unsubscribe(self, handler):
        queue_name = handler.queue_name
        try:
            consumer = self.consumers_by_queue[queue_name]
        except KeyError:
            raise KeyError('there is no subscription for %r' % queue_name)
        if consumer.handler != handler:
            raise KeyError('%s is not subscribed to %r' % (handler, queue_name))
        consumer.stop()
        del self.consumers_by_queue[queue_name]

    def setup_consumer(self, handler):
        if handler.broadcast:
            queue = self.get_queue(handler.queue_name, auto_delete=True, durable=False)
        else:
            queue = self.get_queue(handler.queue_name, auto_delete=handler.once, durable=False)
        consumer = EventConsumer(self, self.connection, queue, handler, self.exchange, max_retries=self.connect_max_retries)
        self.consumers_by_queue[handler.queue_name] = consumer
        return consumer

    @contextmanager
    def get_connection(self):
        with kombu.pools.connections[self.connection].acquire() as conn:
            conn.ensure_connection(max_retries=self.connect_max_retries)
            logger.debug('connecting to %s', conn.as_uri())
            yield conn

    @staticmethod
    def get_queue(name, **kwargs):
        queue_arguments = kwargs.pop('queue_arguments', {})
        queue_arguments['x-ha-policy'] = 'all'
        return kombu.Queue(name, queue_arguments=queue_arguments, **kwargs)

    @staticmethod
    def safe_declare(conn, queue):
        try:
            queue(conn).declare()
        except conn.connection.channel_errors:
            # XXX(Mouad): Redeclare queue since a race condition may happen
            # when declaring queues in failover situation, more info check:
            # https://bugs.launchpad.net/neutron/+bug/1318721.
            queue(conn).declare()

    def emit(self, event, delay=0):
        producer = self._get_producer(event.evt_type, delay)
        producer.emit(event)

    def _get_producer(self, event_type, delay=0):
        try:
            return self._producers[event_type, delay]
        except KeyError:
            if delay:
                producer = EventProducerWithDelay(
                    delay=int(1000 * delay),
                    exchange=self.waiting_exchange,
                    routing_key=event_type,
                    event_system=self,
                )
            else:
                producer = EventProducer(
                    exchange=self.exchange,
                    routing_key=event_type,
                    event_system=self,
                )
        self._producers[event_type, delay] = producer
        return producer



import abc
import collections
import six


@six.add_metaclass(abc.ABCMeta)
class Metric(object):
    def __init__(self, name, tags=None):
        self._name = name
        self._tags = tags or {}

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    def __repr__(self):
        return '%s(name=%r, tags=%r)' % (
            self.__class__.__name__,
            self._name,
            self._tags,
        )

    __str__ = __repr__


class Callable(Metric):
    def __init__(self, name, func, tags=None):
        super(Callable, self).__init__(name, tags)
        self.func = func

    def __iter__(self):
        yield (self._name, self.func(), self._tags)


class Gauge(Metric):
    def __init__(self, name, value=0, tags=None):
        super(Gauge, self).__init__(name, tags)
        self.value = value

    def set(self, value):
        self.value = value

    def __iter__(self):
        yield self._name, self.value, self._tags


class Generator(object):
    def __init__(self, func):
        self.func = func

    def __iter__(self):
        return self.func()


class Aggregate(object):
    def __init__(self, metrics=(), tags=None):
        self._metrics = list(metrics)
        self._tags = tags or {}

    def add(self, metric):
        self._metrics.append(metric)
        return metric

    def add_tags(self, **tags):
        self._tags.update(tags)

    def __iter__(self):
        for metric in self._metrics:
            for name, value, tags in metric:
                tags.update(self._tags)
                yield name, value, tags


class Counter(Metric):
    def __init__(self, name, tags=None):
        super(Counter, self).__init__(name, tags)
        self._value = 0

    def __iadd__(self, value):
        self._value += value
        return self

    def __iter__(self):
        yield self._name, self._value, self._tags


class TaggedCounter(Metric):
    def __init__(self, name, tags=None):
        super(TaggedCounter, self).__init__(name, tags)
        self._values = collections.Counter()

    def incr(self, _by=1, **tags):
        tags.update(self._tags)
        self._values[frozenset(tags.items())] += _by

    def __iter__(self):
        for tags, count in six.iteritems(self._values):
            yield self._name, count, dict(tags)



import textwrap
import functools
import logging

import six

from lymph.core.components import Component, Componentized, ComponentizedBase
from lymph.core.decorators import rpc, RPCBase
from lymph.core.events import TaskHandler, EventHandler
from lymph.core.monitoring import metrics
from lymph.exceptions import RemoteError, EventHandlerTimeout, Timeout, Nack

import gevent
from gevent.event import AsyncResult


logger = logging.getLogger(__name__)
REQUEST_TIMEOUT = 3  # seconds.


class AsyncResultWrapper(object):
    def __init__(self, container, handler, async_result):
        self.container = container
        self.handler = handler
        self.result = async_result

    def get(self, timeout=REQUEST_TIMEOUT):
        try:
            return self.result.get(timeout=timeout)
        except gevent.Timeout:
            self.container.unsubscribe(self.handler)
            raise EventHandlerTimeout


class InterfaceBase(ComponentizedBase):
    def __new__(cls, clsname, bases, attrs):
        methods = {}
        event_handlers = {}
        is_worker = False
        for base in bases:
            if isinstance(base, InterfaceBase):
                methods.update(base.methods)
                is_worker |= base.worker
        for name, value in six.iteritems(attrs):
            if isinstance(value, RPCBase):
                methods[name] = value
            if issubclass(getattr(value, 'cls', object), EventHandler):
                event_handlers[name] = value
            if issubclass(getattr(value, 'cls', object), TaskHandler):
                is_worker = True
        new_cls = super(InterfaceBase, cls).__new__(cls, clsname, bases, attrs)
        new_cls.methods = methods
        new_cls.event_handlers = event_handlers
        new_cls.worker = is_worker
        return new_cls


class ProxyMethod(object):
    def __init__(self, proxy, func):
        self.proxy = proxy
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def defer(self, *args, **kwargs):
        result = AsyncResult()
        self.proxy.spawn(self, *args, **kwargs).link(result)
        return result


class Proxy(Component):
    def __init__(self, container, address, timeout=REQUEST_TIMEOUT, namespace='', error_map=None):
        super(Proxy, self).__init__()
        self._container = container
        self._address = address
        self._method_cache = {}
        self._timeout = timeout
        self._namespace = namespace or address
        self._error_map = error_map or {}
    
    def on_start(self):
        super(Proxy, self).on_start()
        self.timeout_counts = self.metrics.add(metrics.Counter('rpc.timeout_count', {'address': self._address}))
        self.exception_counts = self.metrics.add(metrics.TaggedCounter('rpc.exception_count', {'address': self._address}))

    def _call(self, __name, **kwargs):
        channel = self._container.send_request(self._address, __name, kwargs)
        try:
            return channel.get(timeout=self._timeout).body
        except RemoteError as e:
            error_type = str(e.__class__)
            self.exception_counts.incr(name=e.__class__.__name__)
            if error_type in self._error_map:
                raise self._error_map[error_type]()
            raise
        except Timeout:
            self.timeout_counts += 1
            raise
        except Nack:
            self.exception_counts.incr(name='nack')
            raise

    def __getattr__(self, name):
        try:
            return self._method_cache[name]
        except KeyError:
            method = ProxyMethod(self, functools.partial(self._call, '%s.%s' % (self._namespace, name)))
            self._method_cache[name] = method
            return method


@six.add_metaclass(InterfaceBase)
class Interface(Componentized):
    def __init__(self, container, name=None, builtin=False):
        super(Interface, self).__init__()
        self.container = container
        self.name = name or self.__class__.__name__
        self.base_name = self.name
        self.builtin = builtin
        if container.worker and not builtin:
            self.name = '%s.worker' % self.name

    def should_register(self):
        return True

    def should_install(self):
        return not self.container.worker or self.worker or self.builtin

    def apply_config(self, config):
        self.config = config

    def get_description(self):
        return {}

    def handle_request(self, func_name, channel):
        self.methods[func_name].rpc_call(self, channel, **channel.request.body)

    def request(self, address, subject, body, timeout=REQUEST_TIMEOUT):
        channel = self.container.send_request(address, subject, body)
        return channel.get(timeout=timeout)

    def emit(self, event_type, payload, delay=0):
        self.container.emit_event(event_type, payload, delay=delay)

    def proxy(self, address, **kwargs):
        proxy = Proxy(self.container, address, **kwargs)
        self.add_component(proxy)
        return proxy

    def subscribe(self, *event_types, **kwargs):
        def decorator(func):
            from lymph.core.events import EventHandler
            handler = EventHandler(self, func, event_types, **kwargs)
            self.container.subscribe(handler)
            return handler
        return decorator

    def get_next_event(self, *event_types, **kwargs):
        kwargs['once'] = True
        result = AsyncResult()

        @self.subscribe(*event_types, **kwargs)
        def handler(interface, event):
            result.set(event.body)

        return AsyncResultWrapper(self.container, handler, result)


class DefaultInterface(Interface):
    @rpc()
    def ping(self, payload=None):
        return payload

    @rpc()
    def status(self):
        return {
            'endpoint': self.container.endpoint,
            'identity': self.container.identity,
        }

    @rpc()
    def inspect(self):
        """
        Returns a description of all available rpc methods of this service
        """
        methods = []
        for interface_name, interface in list(self.container.installed_interfaces.items()):
            for name, func in six.iteritems(interface.methods):
                methods.append({
                    'name': '%s.%s' % (interface_name, name),
                    'params': list(func.args.args),
                    'help': textwrap.dedent(func.__doc__ or '').strip(),
                })
        return {
            'methods': methods,
        }

    @rpc()
    def change_loglevel(self, qualname, loglevel, period=60):
        curr_logger = logging.getLogger(qualname)

        def reset():
            logger.info("Resetting logger %s level to %s", qualname, self.container.loglevel)
            curr_logger.setLevel(self.container.loglevel)

        gevent.spawn_later(period, reset)
        curr_logger.setLevel(loglevel)
        logger.info("Changing logger %s level to %s", qualname, loglevel)

    @rpc()
    def get_metrics(self):
        return list(self.container.metrics)



import json

from lymph.client import Client
from lymph.cli.base import Command
from lymph.core import trace


class EmitCommand(Command):
    """
    Usage: lymph emit <event-type> [<body>] [options]

    Emits an event in the event system

    Options:
      --trace-id=<trace_id>        Use the given trace_id.

    {COMMON_OPTIONS}
    """

    short_description = 'Emits an event in the event system'

    def run(self):
        event_type = self.args.get('<event-type>')
        body = json.loads(self.args.get('<body>'))

        trace.set_id(self.args.get('--trace-id'))
        client = Client.from_config(self.config)
        client.emit(event_type, body)



import operator
import unittest

from lymph.core.monitoring import metrics


class CounterMetricsTest(unittest.TestCase):

    def test_get(self):
        counter = metrics.Counter('requests')

        self.assertEqual(list(counter), [('requests', 0, {})])

    def test_repr(self):
        counter = metrics.Counter('requests')

        self.assertEqual(str(counter), repr(counter))
        self.assertEqual(repr(counter), "Counter(name='requests', tags={})")

    def test_increment_by_one(self):
        counter = metrics.Counter('requests')
        counter += 1

        self.assertEqual(list(counter), [('requests', 1, {})])

    def test_increment_by_many(self):
        counter = metrics.Counter('requests')
        counter += 66

        self.assertEqual(list(counter), [('requests', 66, {})])


class TaggedCounterMetricsTest(unittest.TestCase):

    def test_incr_one_type(self):
        error_counter = metrics.TaggedCounter('exception')

        error_counter.incr(type='ValueError')
        error_counter.incr(4, type='ValueError')

        self.assertEqual(list(error_counter), [
            ('exception', 5, {'type': 'ValueError'})])

    def test_incr_different_types(self):
        error_counter = metrics.TaggedCounter('exception')

        error_counter.incr(type='ValueError')
        error_counter.incr(type='ValueError')

        error_counter.incr(type='Nack')

        self.assertEqual(
            sorted(error_counter, key=operator.itemgetter(1)), [
                ('exception', 1, {'type': 'Nack'}),
                ('exception', 2, {'type': 'ValueError'}),
            ])


class AggregateMetricsTest(unittest.TestCase):
    def test_aggregate(self):
        agg = metrics.Aggregate([
            metrics.Gauge('a', 1, tags={'x': '1'}),
            metrics.Gauge('b', 2)
        ], tags={'y': '2'})

        self.assertEqual(sorted(list(agg)), [
            ('a', 1, {'x': '1', 'y': '2'}),
            ('b', 2, {'y': '2'}),
        ])


class GeneratorMetricsTest(unittest.TestCase):
    def test_generator(self):
        def get_metrics():
            yield 'name', 42, {}
            yield 'name', 41, {'x': '1'}

        agg = metrics.Generator(get_metrics)
        expected = [
            ('name', 42, {}),
            ('name', 41, {'x': '1'}),
        ]
        self.assertEqual(list(agg), expected)
        self.assertEqual(list(agg), expected)



from lymph.core.monitoring.metrics import Aggregate
from lymph.core.monitoring.global_metrics import RUsageMetrics, GeventMetrics, GarbageCollectionMetrics


class Aggregator(Aggregate):
    @classmethod
    def from_config(cls, config):
        tags = config.get_raw('tags', {})
        # FIXME: move default metrics out
        return cls([
            RUsageMetrics(),
            GarbageCollectionMetrics(),
            GeventMetrics(),
        ], **tags)



import unittest

from lymph.core.monitoring.metrics import Gauge
from lymph.core.monitoring.aggregator import Aggregator


class AggregatorTestCase(unittest.TestCase):

    def test_aggregator_one_component(self):
        aggr = Aggregator([Gauge('dummy', 'one')])

        self.assertIn(('dummy', 'one', {}), list(aggr))

    def test_aggregator_multiple_components(self):
        aggr = Aggregator([
            Gauge('dummy', 'one'),
            Gauge('dummy', 'two')
        ], tags={'name': 'test'})

        self.assertIn(('dummy', 'one', {'name': 'test'}), list(aggr))
        self.assertIn(('dummy', 'two', {'name': 'test'}), list(aggr))

    def test_aggregator_add_multiple_tags(self):
        aggr = Aggregator([
            Gauge('dummy', 'one')
        ], tags={'name': 'test'})

        aggr.add_tags(origin='localhost', time='now')

        self.assertIn(('dummy', 'one', {'name': 'test', 'origin': 'localhost', 'time': 'now'}), list(aggr))
